"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/*
   Copyright 2016 Yuki KAN

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/
const fs = require("fs");
const http = require("http");
const querystring = require("querystring");
const yaml = require("js-yaml");
const pkg = require("../package.json");
const spec = yaml.safeLoad(fs.readFileSync(__dirname + "/../api.yml", "utf8"));
class Client {
    constructor() {
        this.basePath = spec.basePath;
        /** positive integer */
        this.priority = 0;
        this.host = "";
        this.port = 40772;
        this.socketPath = process.platform === "win32" ? "\\\\.\\pipe\\mirakurun" : "/var/run/mirakurun.sock";
        /** provide User-Agent string to identify client. */
        this.userAgent = "";
        this._userAgent = `MirakurunClient/${pkg.version} Node/${process.version} (${process.platform})`;
    }
    request(method, path, option = {}) {
        return new Promise((resolve, reject) => {
            this._httpRequest(method, path, option).then(res => {
                const ret = {
                    status: res.statusCode,
                    statusText: res.statusMessage,
                    contentType: res.headers["content-type"].split(";")[0],
                    headers: res.headers,
                    isSuccess: (res.statusCode >= 200 && res.statusCode <= 202)
                };
                const chunks = [];
                res.on("data", chunk => chunks.push(chunk));
                res.on("end", () => {
                    const buffer = Buffer.concat(chunks);
                    if (ret.contentType === "application/json") {
                        ret.body = JSON.parse(buffer.toString("utf8"));
                    }
                    else {
                        ret.body = buffer;
                    }
                    if (ret.isSuccess === true) {
                        resolve(ret);
                    }
                    else {
                        reject(ret);
                    }
                });
            }, err => {
                const ret = {
                    status: -1,
                    statusText: "Request Failure",
                    contentType: "",
                    headers: {},
                    isSuccess: false,
                    body: err
                };
                reject(ret);
            });
        });
    }
    getChannels(query) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/channels", { query: query });
            return res.body;
        });
    }
    getChannelsByType(type, query) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/channels/${type}`, { query: query });
            return res.body;
        });
    }
    getChannel(type, channel) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/channels/${type}/${channel}`);
            return res.body;
        });
    }
    getServicesByChannel(type, channel) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/channels/${type}/${channel}/services`);
            return res.body;
        });
    }
    getServiceByChannel(type, channel, sid) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/channels/${type}/${channel}/services/${sid}`);
            return res.body;
        });
    }
    getServiceStreamByChannel(type, channel, sid, decode) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._getTS(`/channels/${type}/${channel}/services/${sid}/stream`, decode);
        });
    }
    getChannelStream(type, channel, decode) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._getTS(`/channels/${type}/${channel}/stream`, decode);
        });
    }
    getPrograms(query) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/programs", { query: query });
            return res.body;
        });
    }
    getProgram(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/programs/${id}`);
            return res.body;
        });
    }
    getProgramStream(id, decode) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._getTS(`/programs/${id}/stream`, decode);
        });
    }
    getServices(query) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/services", { query: query });
            return res.body;
        });
    }
    getService(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/services/${id}`);
            return res.body;
        });
    }
    getLogoImage(id) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/services/${id}/logo`);
            return res.body;
        });
    }
    getServiceStream(id, decode) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._getTS(`/services/${id}/stream`, decode);
        });
    }
    getTuners() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/tuners");
            return res.body;
        });
    }
    getTuner(index) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/tuners/${index}`);
            return res.body;
        });
    }
    getTunerProcess(index) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", `/tuners/${index}/process`);
            return res.body;
        });
    }
    killTunerProcess(index) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("DELETE", `/tuners/${index}/process`);
            return res.body;
        });
    }
    getEvents() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/events");
            return res.body;
        });
    }
    getEventsStream(query) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._requestStream("GET", "/events/stream", { query: query });
        });
    }
    getChannelsConfig() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/config/channels");
            return res.body;
        });
    }
    updateChannelsConfig(channels) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("PUT", "/config/channels", { body: channels });
            return res.body;
        });
    }
    channelScan(option) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._requestStream("PUT", "/config/channels/scan", {
                query: option
            });
        });
    }
    getServerConfig() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/config/server");
            return res.body;
        });
    }
    updateServerConfig(server) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("PUT", "/config/server", { body: server });
            return res.body;
        });
    }
    getTunersConfig() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/config/tuners");
            return res.body;
        });
    }
    updateTunersConfig(tuners) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("PUT", "/config/tuners", { body: tuners });
            return res.body;
        });
    }
    getLog() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/log");
            return res.body;
        });
    }
    getLogStream() {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._requestStream("GET", "/log/stream");
        });
    }
    checkVersion() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/version");
            return res.body;
        });
    }
    updateVersion(force) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this._requestStream("PUT", "/version/update", { query: { force: force } });
        });
    }
    getStatus() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("GET", "/status");
            return res.body;
        });
    }
    restart() {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this.request("PUT", "/restart");
            return res.body;
        });
    }
    _httpRequest(method, path, option = {}) {
        const opt = {
            method: method,
            path: this.basePath + path,
            headers: option.headers || {},
            agent: this.agent
        };
        if (this.host === "") {
            opt.socketPath = this.socketPath;
        }
        else {
            opt.host = this.host;
            opt.port = this.port;
        }
        if (this.userAgent === "") {
            opt.headers["User-Agent"] = this._userAgent;
        }
        else {
            opt.headers["User-Agent"] = this.userAgent + " " + this._userAgent;
        }
        if (option.priority === undefined) {
            option.priority = this.priority;
        }
        opt.headers["X-Mirakurun-Priority"] = option.priority.toString(10);
        if (typeof option.query === "object") {
            opt.path += "?" + querystring.stringify(option.query);
        }
        if (typeof option.body === "object") {
            opt.headers["Content-Type"] = "application/json; charset=utf-8";
            option.body = JSON.stringify(option.body);
        }
        return new Promise((resolve, reject) => {
            const req = http.request(opt, res => {
                if (res.statusCode > 300 && res.statusCode < 400 && res.headers.location) {
                    if (/^\//.test(res.headers.location) === false) {
                        reject(new Error(`Error: Redirecting location "${res.headers.location}" isn't supported.`));
                        return;
                    }
                    this._httpRequest(method, res.headers.location, option)
                        .then(resolve, reject);
                    return;
                }
                resolve(res);
            });
            req.on("error", reject);
            // write request body
            if (typeof option.body === "string") {
                req.write(option.body + "\n");
            }
            req.end();
        });
    }
    _requestStream(method, path, option = {}) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield this._httpRequest(method, path, option);
            if (res.statusCode >= 200 && res.statusCode <= 202) {
                return res;
            }
            else {
                throw res;
            }
        });
    }
    _getTS(path, decode = true) {
        return __awaiter(this, void 0, void 0, function* () {
            const option = {
                query: {
                    decode: decode ? "1" : "0"
                }
            };
            const res = yield this._requestStream("GET", path, option);
            if (res.headers["content-type"] === "video/MP2T") {
                return res;
            }
            else {
                throw res;
            }
        });
    }
}
exports.default = Client;

//# sourceMappingURL=client.js.map
